package serie_junit;

/**
 *  Un importo in denaro.
 */
public class Importo {

    private final long _cent;

    /**
     *  Costruttore per l'oggetto Importo
     *
     *@param  positivo  True se il importo  in ingresso, false se  in uscita.
     *@param  euro      Importo in Euro (esclusi centesimi) - sempre positivo.
     *@param  cent      Centesimi - sempre positivo.
     */
    public Importo(boolean positivo, long euro, long cent) {
        if(euro < 0 || cent < 0)
            throw new NumberFormatException("Valori negativi nel costruttore di un Importo");
        long centesimiSenzaSegno = euro * 100 + cent;
        if(positivo)
            _cent = centesimiSenzaSegno;
        else
            _cent = -centesimiSenzaSegno;
    }

    /**
     *  Costruttore per l'oggetto Importo
     *
     *@param  cent      Il numero totale di centesimi, positivo o negativo.
     */
    public Importo(long cent) {
        _cent = cent;
    }

    /**
     *  Restituisce l'importo come una stringa.
     *
     * @return    Il valore dell'importo.
     */
    public String getValore() {
        long euro = Math.abs((long)(_cent / 100));
        long centesimi = Math.abs(_cent % 100);

        String risultato = "";
        if(_cent < 0)
            risultato += "-";
        risultato = risultato + euro + ".";
        if(centesimi < 10)
            risultato += "0";
        risultato += centesimi;

        return risultato;
    }

   /**
     *  Restituisce la somma dell'importo con un altro Importo.
     *
     * @param    L'altro importo.
     * @return   L'importo somma.
     */
    public Importo somma(Importo i) {
        return new Importo(this._cent + i._cent);
    }
}
