// -------------------------------------------------------------
// Primitive
// scritto da Stefano Cristiano (stefano.cristiano@email.it)
//
// Questo esempio si propone di spiegare il funzionamento 
// delle primitive di disegno di Direct3D, attraverso l'utilizzo
// di vertex ed index buffer
// Un ringraziamento speciale a Luciano Iurino dei PM Studios di Bari 
// (http://www.pmstudios.it) per aver concesso lutilizzo di un suo modello dal 
// gioco in lavorazione ETROM (http://www.etrom.net) per lesempio sul per
// pixel lighing.
// -------------------------------------------------------------



#include "stdafx.h"
#include <windows.h>
#include <d3dx9.h>
#include <atlbase.h>
#include <string>
#include <vector>
//Questo  un metodo per linkare le librerie
//di DX9
#pragma comment(lib,"d3d9.lib")
#pragma comment(lib,"d3dx9.lib")

//Typedefiamo questi "smart refcounted pointers"
typedef CComPtr<IDirect3D9>			Direct3D9_Ptr;
typedef CComPtr<IDirect3DDevice9>	Direct3DDevice9_Ptr;
typedef CComPtr<ID3DXFont>			D3DXFont_Ptr;
typedef CComPtr<ID3DXMesh>			D3DXMesh_Ptr;
typedef CComPtr<IDirect3DTexture9>	Direct3DTexture9_Ptr;
typedef CComPtr<ID3DXEffect>		D3DXEffect_Ptr;
typedef CComPtr<ID3DXBuffer>		D3DXBuffer_Ptr;

HWND				g_hwnd;		//La finestra di rendering		
Direct3D9_Ptr		g_d3d;		//L'oggetto IDirect3D (serve x creare il device)
Direct3DDevice9_Ptr g_device;	//Il Device (il nostro acceleratore)
D3DXFont_Ptr		g_font;		//Il font per il testo
RECT				g_font_rect;//il rettangolo della finestra
int					g_width;	//Larghezza corrente della finestra
int					g_height;	//Altezza corrente della finestra
bool				g_terminated;//Il programma e' stato terminato?	
bool				g_bexp=true;

bool CreateDeviceObjects();
void Draw();
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);


int APIENTRY WinMain(HINSTANCE instance,HINSTANCE,LPSTR,int)
{
	//Registriamo la WNDCLASS
	const WNDCLASS wndclass = {
			0, WndProc, 0,0, instance,0,
			LoadCursor(NULL, IDC_CROSS),
			0,0,_T("PerPixelLighingClass")
	};
	g_width = 640;
	g_height= 480;
	g_terminated = false;
	if(!RegisterClass(&wndclass))
		return -1;

	//Creiamo una normalissima finestra
	g_hwnd = CreateWindow(	_T("PerPixelLighingClass"),
							_T("Per Pixel Lighing! Con spazio cambi da dst2 a exp! Premere un tasto per uscire."),
							WS_OVERLAPPEDWINDOW,CW_USEDEFAULT,CW_USEDEFAULT,
							g_width,g_height,GetDesktopWindow(),0,instance,0);
	if(!g_hwnd)
		return -1;

	//Creiamo l'oggetto IDirect3D9 (da  notare
	//che assegnamo direttamente il puntatore ricevuto
	//al membro .p dello smart pointer...il perch
	//e' spiegato nell'articolo che spiega il COM)
	g_d3d.p = Direct3DCreate9(D3D_SDK_VERSION);
	if(!g_d3d)
		return -1;

	//Prendiamo la modalit di display corrente del 
	//desktop
	D3DDISPLAYMODE display;
	g_d3d->GetAdapterDisplayMode(D3DADAPTER_DEFAULT,&display);

	//Impostiamo la struttura per un device in finestra,
	//con un solo backbuffer,senza multisampling (antialiasing),
	//e abilitiamo anche lo zbuffer (D3DFMT_D16)
	D3DPRESENT_PARAMETERS params;
	ZeroMemory(&params,sizeof(params));
	params.Windowed					= TRUE;
	params.hDeviceWindow			= g_hwnd;
	params.BackBufferCount			= 1;
	params.BackBufferFormat			= display.Format;
	params.BackBufferWidth			= 640;	
	params.BackBufferHeight			= 480;	
	params.SwapEffect				= D3DSWAPEFFECT_DISCARD;
	params.MultiSampleType			= D3DMULTISAMPLE_NONE;
	params.EnableAutoDepthStencil	= TRUE;
	params.AutoDepthStencilFormat	= D3DFMT_D16;


	//Creiamo il device con accelereazione (D3DDEVTYPE_HAL)
	//usando il software vertex processing (disabilitiamo il T&L
	//della scheda praticamente)
	HRESULT hr;
	hr = g_d3d->CreateDevice(	D3DADAPTER_DEFAULT,D3DDEVTYPE_HAL,
								g_hwnd,D3DCREATE_SOFTWARE_VERTEXPROCESSING,
								&params,&g_device.p);
	if(FAILED(hr))
		return -1;

	//Mostriamo la finestra
	ShowWindow(g_hwnd,SW_SHOW);

	//Aggiorniamo la finestra
	UpdateWindow(g_hwnd);

	//Ne prendiamo le dimensioni
	GetClientRect(g_hwnd,&g_font_rect);

	if(!CreateDeviceObjects())
	{
		//Fine del programma
		MessageBox(g_hwnd,"Errore","Errore di inizializzazione",MB_OK);
		DestroyWindow(g_hwnd);
		return -1;
	}
	//Pompa dei messaggi
	while(!g_terminated)
	{	
		Draw();
		//Smistiamo i messaggi in coda...
		MSG msg;
		while(PeekMessage(&msg,0,0,0,PM_REMOVE))
		{	
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	
	//Fine del programma
	DestroyWindow(g_hwnd);
	return 0;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{

	LRESULT res = 0;
	if(message==WM_CLOSE||message==WM_KEYDOWN)
	{
		if(wParam!=VK_SPACE)
		{	
			g_terminated = true;
			PostQuitMessage(0);
			return 0;
		}else
		{		
			g_bexp = !g_bexp;
		}
	}
	res = DefWindowProc(hWnd,message,wParam,lParam);
	return res;
}


bool LoadMesh(const std::string filename, D3DXMesh_Ptr& mesh, 
				std::vector<D3DXMATERIAL>& materials,
				std::vector<Direct3DTexture9_Ptr>& textures)
{
	DWORD dwNumMaterials;
	HRESULT hr;
	D3DXBuffer_Ptr pD3DXMtrlBuffer;//Per ottenere i materiali
	hr = D3DXLoadMeshFromX( 
			filename.c_str(),		//Nome del file
			D3DXMESH_MANAGED,		//Si gestisce da solo in caso di "Lost Device"
			g_device, NULL,			//Il device e l'adiacenza che non ci interessa
			&pD3DXMtrlBuffer,NULL,	//Ritorna array di materiali (D3DXMATERIAL)
			&dwNumMaterials,		//Ritorna il numero dei material presenti
			&mesh); 				//Ritorna la mesh

	if(FAILED(hr))
		return false;

	//Estraiamo il path dove si trova il file .x
	std::string path = filename.substr(0,filename.find_last_of(_T("\\"))+1);	
	D3DXMATERIAL* d3dxMaterials = (D3DXMATERIAL*)pD3DXMtrlBuffer->GetBufferPointer();		
	for(DWORD i=0; i<dwNumMaterials; i++ )
	{	
		materials.push_back(d3dxMaterials[i]);
		if(!d3dxMaterials[i].pTextureFilename)
		{	
			textures.push_back(0);
			continue;
		}
		std::string texName =d3dxMaterials[i].pTextureFilename;
		Direct3DTexture9_Ptr tex;
		//Proviamo il nome assoluto del file
		hr = D3DXCreateTextureFromFile(g_device,texName.c_str(),&tex);
		if(FAILED(hr))
		{
			//cerchiamo nella dir corrente
			int start = texName.find_last_of('\\');
			texName = texName.substr(start+1,texName.size()-start-1);
			hr = D3DXCreateTextureFromFile(g_device,texName.c_str(),&tex);
			if(FAILED(hr))
			{
				//Cerchiamo nel path del file .x
				texName = path + texName;
				D3DXCreateTextureFromFile(g_device,texName.c_str(),&tex);
			}
		}
		textures.push_back(tex);
	}
	return true;
}

Direct3DTexture9_Ptr att1D,att2DExp,att2D;
std::vector<Direct3DTexture9_Ptr> g_textures;
std::vector<D3DXMATERIAL>		g_materials;
D3DXMesh_Ptr		g_mesh;		//La mesh
D3DXMesh_Ptr		g_lightmesh;//Mesh Della luce
D3DXEffect_Ptr		g_effect;	//L'effect file
D3DXVECTOR3			g_lightPos;	
float				g_lightRadius;

bool CreateDeviceObjects()
{
	HRESULT hr;

	//Controlliamo che l'hardware corrente supporta
	//i vertex e pixel shader 1.1
	D3DCAPS9 caps;
	g_device->GetDeviceCaps(&caps);
	if(!( caps.VertexShaderVersion >= D3DVS_VERSION(1,1) )||
		!( caps.PixelShaderVersion >= D3DPS_VERSION(1,1) ))
	{
		MessageBox(g_hwnd,
					"Hai bisogno di una Geforce3 o superiore "
					"per vedere questo esempio.\nComprala,costa poco :P asdasd", 
					"Errore",MB_OK);
			return false;
	}
	//Istanziamo l'effect file
	hr = D3DXCreateEffectFromFile(g_device,_T("PerPixelLighing.fx"),0,0,0,0,&g_effect,0);
	if(FAILED(hr))
		return false;

	//Abilitiamo il dithering, lo zbuffer ed il calcolo
	//dell'illuminazione
	g_device->SetRenderState(D3DRS_DITHERENABLE, TRUE );
	g_device->SetRenderState(D3DRS_ZENABLE, TRUE );
	g_device->SetRenderState(D3DRS_LIGHTING,FALSE);

	//D3D deve disegnare i poligoni la cui normale 
	//e' rivolta verso la camera, usando il verso antiorario
	//dei vertici x determinare dove punta
	g_device->SetRenderState(D3DRS_CULLMODE,D3DCULL_CCW);

	//Moduliam la texture con il colore dell'illuminazione
	g_device->SetTextureStageState(0,D3DTSS_COLOROP,D3DTOP_MODULATE);
	g_device->SetTextureStageState(0,D3DTSS_COLORARG1,D3DTA_TEXTURE);
	g_device->SetTextureStageState(0,D3DTSS_COLORARG2,D3DTA_CURRENT);

	//Abilitiamo il filtering
	g_device->SetSamplerState(0,D3DSAMP_MINFILTER,D3DTEXF_LINEAR);
	g_device->SetSamplerState(0,D3DSAMP_MAGFILTER,D3DTEXF_LINEAR);
	g_device->SetSamplerState(0,D3DSAMP_MIPFILTER,D3DTEXF_LINEAR);

	//Disabilitiamo gli stages successivi
	g_device->SetTextureStageState(1,D3DTSS_COLOROP,D3DTOP_DISABLE);
	g_device->SetTextureStageState(1,D3DTSS_ALPHAOP,D3DTOP_DISABLE);

	//Impostiamo la luce
	D3DLIGHT9 theLight;
	ZeroMemory(&theLight,sizeof(theLight));
	theLight.Type = D3DLIGHT_DIRECTIONAL;
	theLight.Direction = D3DXVECTOR3(-0.2f, -0.3f, 0.5f);
	theLight.Ambient  = D3DXCOLOR(0.3f,0.3f,0.3f,1.0f);
	theLight.Diffuse  = D3DXCOLOR(1,1,1,1.0f);
	g_device->SetLight(0,&theLight);
	g_device->LightEnable(0,TRUE);
	//Impostiamo il materiale
	D3DMATERIAL9 m;
	ZeroMemory(&m,sizeof(m));
	m.Ambient = D3DXCOLOR(0.3f,0.3f,0.3f,1.0f);
	m.Diffuse = D3DXCOLOR(1,1,1,1);
	g_device->SetMaterial(&m);	
	g_device->SetRenderState(D3DRS_SPECULARENABLE, FALSE);

	if(!LoadMesh(_T("stanza.X"),g_mesh,
		g_materials,g_textures))
		return false;
	D3DXCreateSphere(g_device,0.1f,10,10,&g_lightmesh,0);
	g_lightPos = D3DXVECTOR3(-1,0,1);
	g_lightRadius = 3;

	int size = 64;

	D3DXCreateTexture(g_device,size,size,0,D3DX_DEFAULT,D3DFMT_L8,D3DPOOL_MANAGED,&att2D);
	D3DXCreateTexture(g_device,size,size,0,D3DX_DEFAULT,D3DFMT_L8,D3DPOOL_MANAGED,&att2DExp);
	D3DLOCKED_RECT r_dst2,r_exp;
	hr = att2D->LockRect(0,&r_dst2,0,D3DLOCK_NOSYSLOCK);
	if(FAILED(hr))
		return false;
	hr = att2DExp->LockRect(0,&r_exp,0,D3DLOCK_NOSYSLOCK);
	if(FAILED(hr))
		return false;
	float scale = 1.8f;
	for(int y = 0; y < size; y++)
	{
		float fy = (2*y+1)/(float)size-1;
		float efy = fy * scale;
		efy = exp(-(efy*efy));
		for(int i = 0; i < size; i++)
		{	
			BYTE* D2	= (BYTE*)r_dst2.pBits+r_dst2.Pitch*i+y;
			BYTE* Exp	= (BYTE*)r_exp.pBits +r_exp.Pitch*i +y;
			float fi = (2*i+1)/(float)size-1;
			float efi = fi * scale;
			efi = exp(-efi*efi);
			*Exp = efi*efy*255;
			float dst2 = fi * fi + fy * fy;
			if(dst2 > 1) 
				dst2 = 1;
			*D2 = dst2*255;
		}
	}
	att2D->UnlockRect(0);
	att2DExp->UnlockRect(0);
	D3DXCreateFont(g_device,(HFONT)GetStockObject(SYSTEM_FONT),&g_font.p);
	return true;
}


void Draw()
{
	//Il nostro parametro temporale
	float t = (float)GetTickCount()/1000.0f;
	//Inizio del frame corrente
	g_device->BeginScene();
	//Puliamo lo schermo con 1 bel nero
	g_device->Clear(0,0,D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,0,1,0);

	//Questo sarebbe il rapporto tra le dimensioni dello schermo
	float aspect = (float)g_width/(float)g_height;
	D3DXMATRIX matProj,matView,matWorld;
	D3DXMatrixIdentity(&matWorld);
	//La matrice di proiezione
	D3DXMatrixPerspectiveFovLH(&matProj,D3DX_PI/3.0f,aspect,0.1f,50.0f);
	//La matrice camera (Fissa)
	D3DXMatrixLookAtLH(&matView,
						&D3DXVECTOR3(8,15,8),
//						&D3DXVECTOR3(3,5,-3),
						&D3DXVECTOR3(0,0,0),
						&D3DXVECTOR3(0,1,0));
	//animiamo la luce
	g_lightPos.x = sinf(t)*7;
	g_lightPos.z = cosf(t)*7;
	g_lightPos.y = cosf(2*t)*3+3;
	g_lightRadius = 5;


	g_device->SetTransform(D3DTS_WORLD,&matWorld);
	g_device->SetTransform(D3DTS_VIEW,&matView);
	g_device->SetTransform(D3DTS_PROJECTION,&matProj);
	// set light direction
	g_effect->SetVector("light", &D3DXVECTOR4(g_lightPos.x,g_lightPos.y,g_lightPos.z,g_lightRadius));
	D3DXMATRIX mWorldViewProj = matWorld * matView * matProj;
	g_effect->SetMatrix( "matWorldViewProj", &mWorldViewProj );
	g_effect->SetMatrix( "matWorld", &matWorld);
	D3DXHANDLE tec;
	if(g_bexp)
	{
		g_effect->SetTexture("att1DTexture", att2DExp);
		g_effect->SetTexture("att2DTexture", att2DExp);	
		tec = g_effect->GetTechniqueByName("TecnicaEXP2");
	}else
	{
		tec = g_effect->GetTechniqueByName("TecnicaDST2");
		g_effect->SetTexture("att1DTexture", att2D);
		g_effect->SetTexture("att2DTexture", att2D);	
	}

	if(SUCCEEDED(g_effect->ValidateTechnique(tec)))
		g_effect->SetTechnique(tec);
	else
		return;
	
	for(UINT i = 0; i < g_materials.size(); i++)
	{
		UINT passes = 0;
		g_effect->SetTexture("diffuseTexture",g_textures[i]);
		g_effect->Begin(&passes,0);
		for(UINT p = 0; p < passes; p++)
		{
			g_effect->Pass(p);
			//Disegnamo la mesh
			g_mesh->DrawSubset(i);
		}
		g_effect->End();
	}

	
	g_device->SetRenderState(D3DRS_LIGHTING,FALSE);
	D3DXMatrixTranslation(&matWorld,g_lightPos.x,g_lightPos.y,g_lightPos.z);
	g_device->SetTransform(D3DTS_WORLD,&matWorld);
	g_device->SetTexture(0,0);
	g_lightmesh->DrawSubset(0);	
	g_font->DrawText("Con SPAZIO passi da exp(att^2) a dst(att^2)",-1,&g_font_rect,0,
					0Xffffffff);
	//fine del frame
	g_device->EndScene();
	//riversiamo a schermo il tutto
	g_device->Present(0,0,0,0);
}
