// -------------------------------------------------------------
// PerPixelLighing.fx
//
// Calcola e applica una mappa di attenuazione x l'illuminazione
// scritto da Stefano Cristiano (stefano.cristiano@email.it)
// -------------------------------------------------------------


// -------------------------------------------------------------
// Parametri passati agli shaders
// -------------------------------------------------------------
float4x4 matWorldViewProj;		//Concatenata di matrice world,view e proj
float4x4 matWorld;				//Matrice world
float4 light;					//Posizione luce in world space
texture diffuseTexture;			//texture diffusa
texture att1DTexture;			//Texture di attenuazione 1
texture att2DTexture;			//Texture di attenuazione 2
float4 ambient = {0.1,0.1,0.1,1};//Colore ambient


//Quello che sputa fuori il nostro vertex shader
struct VS_OUTPUT
{
    float4 Pos  : POSITION;
	float2 att2D: TEXCOORD0;
	float2 att1D: TEXCOORD1;
	float2 tex	: TEXCOORD2;
};

// -------------------------------------------------------------
// VERTEX SHADER: genera le coordinate x texture di attenuazione
// -------------------------------------------------------------
VS_OUTPUT VS(float4 Pos : POSITION, float2 diffTexCoord : TEXCOORD0)
{
    VS_OUTPUT Out = (VS_OUTPUT)0;  
	
	//Questo potrebbe anche essere fatto
	//nell'applicazione x ottimizzare...
	light.w *= 2;
							
	//Posizione vertice in world space
    float3 WorldPos			= mul(Pos,matWorld);
	
	//Vettore dalla luce al vertice
	float3 light_to_vertex	= WorldPos-light;

	//Calcoliamo le coordinate x l'attenuazione
    float3 att_texcoord		= light_to_vertex/light.w+0.5;
	
	//Scriviamo la posizione finale
	Out.Pos					= mul(Pos, matWorldViewProj);

	//Scriviamo le coordinate x l'attenuazione
	Out.att2D				= att_texcoord.xy;
	Out.att1D.x				= att_texcoord.z;
	Out.att1D.y				= 0.5f;

	//Copiamo cosi' come sono le coord. x la texture diffusa
	Out.tex					= diffTexCoord;
	return Out;
}


sampler mySampler0 = sampler_state
{
	MinFilter	= LINEAR;			//Filtering trilineare
	MagFilter	= LINEAR;
	MipFilter	= LINEAR;
	AddressU	= CLAMP;			//Non vogliamo che la nostra texture sia ripetuta in tiling...
	AddressV	= CLAMP;
	Texture		= (att2DTexture);	//La texture
};

sampler mySampler1 = sampler_state
{
	MinFilter	= LINEAR;			//Filtering trilineare
	MagFilter	= LINEAR;
	MipFilter	= LINEAR;
	AddressU	= CLAMP;			//Non vogliamo che la nostra texture sia ripetuta in tiling...
	AddressV	= CLAMP;
	Texture		= (att1DTexture);	//La texture
};

sampler mySampler2 = sampler_state
{
	MinFilter	= LINEAR;			//Filtering trilineare
	MagFilter	= LINEAR;
	MipFilter	= LINEAR;
	AddressU	= WRAP;				//vogliamo che la nostra texture sia ripetuta in tiling...
	AddressV	= WRAP;
	Texture		= (diffuseTexture);	//La texture
};
// -------------------------------------------------------------
// Pixel Shader per attenuazione DST^2
// -------------------------------------------------------------
float4 PS(	float2 att2D : TEXCOORD0, 
			float2 att1D : TEXCOORD1, 
			float2 tex : TEXCOORD2) : COLOR
{
	//Calcoliamo la luce che viene con (1-(x/r)2 - (y/r)2 - (z/r)2 + ambient)
	//dove il 2 sta per quadrato. usiamo la saturate x evitare
	//che si esca dal range [0,1]
	float4 spot = (saturate(1-tex2D(mySampler0,att2D)-tex2D(mySampler1,att1D))+ambient);
	
	//Moduliamo la texture diffusa x la luce
	//e moltiplichiamo x 4 x rendere tutto + lucente
	return tex2D(mySampler2,tex)*spot*4;
	
}

float4 PSEXP(float2 att2D : TEXCOORD0, 
			float2 att1D : TEXCOORD1, 
			float2 tex : TEXCOORD2) : COLOR
{
	//Semplice modulazione con saturazione e aggiunta di ambient
	float4 spot = (saturate(tex2D(mySampler0,att2D)*tex2D(mySampler1,att1D))+ambient);

	//Moduliamo la texture diffusa x la luce
	//e moltiplichiamo x 4 x rendere tutto + lucente
	return tex2D(mySampler2,tex)*spot*4;
	
}
// -------------------------------------------------------------
// TECNICA DST^2
// -------------------------------------------------------------
technique TecnicaDST2
{
    pass P0
    {
        // compiler directives
        VertexShader = compile vs_1_1 VS();
        PixelShader  = compile ps_1_1 PS();

    }
}

// -------------------------------------------------------------
// TECNICA EXP^2
// -------------------------------------------------------------
technique TecnicaEXP2
{
    pass P0
    {
        // compiler directives
        VertexShader = compile vs_1_1 VS();
        PixelShader  = compile ps_1_1 PSEXP();

    }
}

