// -------------------------------------------------------------
// Primitive
// scritto da Stefano Cristiano (stefano.cristiano@email.it)
//
// Questo esempio si propone di spiegare il funzionamento 
// delle primitive di disegno di Direct3D, attraverso l'utilizzo
// di vertex ed index buffer
// -------------------------------------------------------------

#include "stdafx.h"
#include <d3dx9.h>
#include <atlbase.h>
#include <vector>
#include <string>
//Questo  un metodo per linkare le librerie
//di DX9
#pragma comment(lib,"d3d9.lib")
#pragma comment(lib,"d3dx9.lib")

//Typedefiamo questi "smart refcounted pointers"
typedef CComPtr<IDirect3D9>			Direct3D9_Ptr;
typedef CComPtr<IDirect3DDevice9>	Direct3DDevice9_Ptr;
typedef CComPtr<ID3DXFont>			D3DXFont_Ptr;
typedef CComPtr<IDirect3DTexture9>	Direct3DTexture9_Ptr;
typedef CComPtr<IDirect3DVertexBuffer9> D3DVBuffer_ptr;
typedef CComPtr<IDirect3DIndexBuffer9> D3DIBuffer_ptr;

HWND				g_hwnd;		//La finestra di rendering		
Direct3D9_Ptr		g_d3d;		//L'oggetto IDirect3D (serve x creare il device)
Direct3DDevice9_Ptr g_device;	//Il Device (il nostro acceleratore)
int					g_width;	//Larghezza corrente della finestra
int					g_height;	//Altezza corrente della finestra
bool				g_terminated;//Il programma e' stato terminato?	
RECT				g_font_rect;
D3DXFont_Ptr		g_font;

bool CreateDeviceObjects();
void Draw();
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);


int APIENTRY WinMain(HINSTANCE instance,HINSTANCE,LPSTR,int)
{
	//Registriamo la WNDCLASS
	const WNDCLASS wndclass = {
			0, WndProc, 0,0, instance,0,
			LoadCursor(NULL, IDC_CROSS),
			0,0,_T("PrimitiveClass")
	};
	g_width = 640;
	g_height= 480;
	g_terminated = false;
	if(!RegisterClass(&wndclass))
		return -1;

	//Creiamo una normalissima finestra
	g_hwnd = CreateWindow(	_T("PrimitiveClass"),
							_T("Primitive!Premere ESC per uscire."),
							WS_OVERLAPPEDWINDOW,CW_USEDEFAULT,CW_USEDEFAULT,
							g_width,g_height,GetDesktopWindow(),0,instance,0);
	if(!g_hwnd)
		return -1;

	//Creiamo l'oggetto IDirect3D9 (da  notare
	//che assegnamo direttamente il puntatore ricevuto
	//al membro .p dello smart pointer...il perch
	//e' spiegato nell'articolo che spiega il COM)
	g_d3d.p = Direct3DCreate9(D3D_SDK_VERSION);
	if(!g_d3d)
		return -1;

	//Prendiamo la modalit di display corrente del 
	//desktop
	D3DDISPLAYMODE display;
	g_d3d->GetAdapterDisplayMode(D3DADAPTER_DEFAULT,&display);

	//Impostiamo la struttura per un device in finestra,
	//con un solo backbuffer,senza multisampling (antialiasing),
	//e abilitiamo anche lo zbuffer (D3DFMT_D16)
	D3DPRESENT_PARAMETERS params;
	ZeroMemory(&params,sizeof(params));
	params.Windowed					= TRUE;
	params.hDeviceWindow			= g_hwnd;
	params.BackBufferCount			= 1;
	params.BackBufferFormat			= display.Format;
	params.BackBufferWidth			= 640;	
	params.BackBufferHeight			= 480;	
	params.SwapEffect				= D3DSWAPEFFECT_DISCARD;
	params.MultiSampleType			= D3DMULTISAMPLE_NONE;
	params.EnableAutoDepthStencil	= TRUE;
	params.AutoDepthStencilFormat	= D3DFMT_D16;


	//Creiamo il device con accelereazione (D3DDEVTYPE_HAL)
	//usando il software vertex processing (disabilitiamo il T&L
	//della scheda praticamente)
	HRESULT hr;
	hr = g_d3d->CreateDevice(	D3DADAPTER_DEFAULT,D3DDEVTYPE_HAL,
								g_hwnd,D3DCREATE_SOFTWARE_VERTEXPROCESSING,
								&params,&g_device.p);
	if(FAILED(hr))
		return -1;

	//Mostriamo la finestra
	ShowWindow(g_hwnd,SW_SHOW);

	//Aggiorniamo la finestra
	UpdateWindow(g_hwnd);

	//Ne prendiamo le dimensioni
	GetClientRect(g_hwnd,&g_font_rect);

	if(!CreateDeviceObjects())
	{
		//Fine del programma
		MessageBox(g_hwnd,"Errore","Errore di inizializzazione",MB_OK);
		DestroyWindow(g_hwnd);
		return -1;
	}
	//Pompa dei messaggi
	while(!g_terminated)
	{	
		Draw();
		//Smistiamo i messaggi in coda...
		MSG msg;
		while(PeekMessage(&msg,0,0,0,PM_REMOVE))
		{	
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	
	//Fine del programma
	DestroyWindow(g_hwnd);
	return 0;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{

	LRESULT res = 0;
	if(message==WM_CLOSE||(message==WM_KEYDOWN&&wParam==VK_ESCAPE))
	{
		g_terminated = true;
		PostQuitMessage(0);
		return 0;
	}
	res = DefWindowProc(hWnd,message,wParam,lParam);
	return res;
}

struct MyVertex
{
	D3DXVECTOR3 pos;
	D3DXVECTOR2 tex;
};
const DWORD MyVertex_FVF = (D3DFVF_XYZ|D3DFVF_TEX1);
D3DVBuffer_ptr vb;
D3DIBuffer_ptr ib;
Direct3DTexture9_Ptr texture;
D3DPRIMITIVETYPE pt;
bool FillBuffers();

bool CreateDeviceObjects()
{
	HRESULT hr;
	//Creiamo il font
	D3DXCreateFont(g_device,(HFONT)GetStockObject(SYSTEM_FONT),&g_font.p);
	
	pt = D3DPT_TRIANGLELIST;
	g_device->CreateVertexBuffer(
		5*sizeof(MyVertex),D3DUSAGE_WRITEONLY,
		MyVertex_FVF,D3DPOOL_MANAGED,&vb,0);
	g_device->CreateIndexBuffer(
		9*sizeof(WORD),D3DUSAGE_WRITEONLY,
		D3DFMT_INDEX16,D3DPOOL_MANAGED,&ib,0);
	hr = D3DXCreateTextureFromFile(g_device,"trianglestrips.png",&texture);
	if(FAILED(hr))
		return false;

	MyVertex* vertices;
	vb->Lock(0,0,(void**)&vertices,D3DLOCK_NOSYSLOCK);
	float w = 1;
	float h = 1;
	vertices[0].pos = D3DXVECTOR3(-w/2,	h/2, 0);
	vertices[0].tex = D3DXVECTOR2(0,0);
	vertices[1].pos = D3DXVECTOR3(0,	h/2, 0);
	vertices[1].tex = D3DXVECTOR2(0.5f,0);
	vertices[2].pos = D3DXVECTOR3(w/2,	h/2, 0);
	vertices[2].tex = D3DXVECTOR2(1,0);
	vertices[3].pos = D3DXVECTOR3(w/2,	-h/2,0);
	vertices[3].tex = D3DXVECTOR2(1,1);
	vertices[4].pos = D3DXVECTOR3(-w/2,	-h/2,0);
	vertices[4].tex = D3DXVECTOR2(0,1);
	vb->Unlock();

	//Abilitiamo il dithering, lo zbuffer
	g_device->SetRenderState(D3DRS_DITHERENABLE, TRUE );
	g_device->SetRenderState(D3DRS_ZENABLE, TRUE );
	g_device->SetRenderState(D3DRS_LIGHTING,FALSE);

	//D3D deve disegnare i poligoni la cui normale 
	//e' rivolta verso la camera, usando il verso antiorario
	//dei vertici x determinare dove punta
	g_device->SetRenderState(D3DRS_CULLMODE,D3DCULL_CCW);

	//Usiamo la texture per il colore dei poligoni
	g_device->SetTextureStageState(0,D3DTSS_COLOROP,D3DTOP_SELECTARG1);
	g_device->SetTextureStageState(0,D3DTSS_COLORARG1,D3DTA_TEXTURE);

	//Abilitiamo il filtering
	g_device->SetSamplerState(0,D3DSAMP_MINFILTER,D3DTEXF_LINEAR);
	g_device->SetSamplerState(0,D3DSAMP_MAGFILTER,D3DTEXF_LINEAR);
	g_device->SetSamplerState(0,D3DSAMP_MIPFILTER,D3DTEXF_LINEAR);

	//Disabilitiamo gli stages successivi
	g_device->SetTextureStageState(1,D3DTSS_COLOROP,D3DTOP_DISABLE);
	g_device->SetTextureStageState(1,D3DTSS_ALPHAOP,D3DTOP_DISABLE);

	//Impostiamo la luce
	D3DLIGHT9 theLight;
	ZeroMemory(&theLight,sizeof(theLight));
	theLight.Type = D3DLIGHT_DIRECTIONAL;
	theLight.Direction = D3DXVECTOR3(1, 1, 1);
	theLight.Ambient  = D3DXCOLOR(0.1f,0.1f,0.1f,1.0f);
	theLight.Diffuse  = D3DXCOLOR(1,1,1,1.0f);
	g_device->SetLight(0,&theLight);
	g_device->LightEnable(0,TRUE);

	//Impostiamo il materiale
	D3DMATERIAL9 m;
	ZeroMemory(&m,sizeof(m));
	m.Ambient = D3DXCOLOR(0.1f,0.1f,0.1f,1.0f);
	m.Diffuse = D3DXCOLOR(1,1,1,1);
	g_device->SetMaterial(&m);	
	g_device->SetRenderState(D3DRS_SPECULARENABLE, FALSE);

	return FillBuffers();
}


bool FillBuffers()
{ 
	WORD* indices;
	ib->Lock(0,0,(void**)&indices, D3DLOCK_NOSYSLOCK);
	WORD i_fan[]	= {1,2,3,4,0};
	WORD i_strip[]	= {0,1,4,2,3};
	WORD i_list[]	= {	0,1,4,	//Triangolo 1
						1,3,4,	//Triangolo 2
						1,2,3};	//Triangolo 3
	switch(pt)
	{
	case D3DPT_TRIANGLEFAN:
		memcpy(indices,i_fan,sizeof(i_fan));
		break;
		
	case D3DPT_TRIANGLESTRIP:
		memcpy(indices,i_strip,sizeof(i_strip));
		break;
		
	case D3DPT_TRIANGLELIST:
		memcpy(indices,i_list,sizeof(i_list));
		break;
	}
	ib->Unlock();
	return true; 
}

const TCHAR* GetPT(D3DPRIMITIVETYPE pt)
{
	switch(pt)
	{
	case D3DPT_TRIANGLEFAN:
		return "D3DPT_TRIANGLEFAN";
		break;
	case D3DPT_TRIANGLESTRIP:
		return "D3DPT_TRIANGLESTRIP";
		break;
	case D3DPT_TRIANGLELIST:
		return "D3DPT_TRIANGLELIST";
		break;
	}
	return "UNDEFINED";
}


void Draw()
{
	//Inizio del frame corrente
	g_device->BeginScene();
	//Puliamo lo schermo con 1 bel nero
	g_device->Clear(0,0,D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,0,1,0);

	D3DXMATRIX matProj;
	//La matrice di proiezione
	D3DXMatrixOrthoLH(&matProj,1,1,-0.1f,1);

	g_device->SetTransform(D3DTS_PROJECTION,&matProj);
	
	g_device->SetTexture(0,texture);
	g_device->SetFVF(MyVertex_FVF);
	g_device->SetStreamSource(0,vb,0,sizeof(MyVertex));
	g_device->SetIndices(ib);
	g_device->DrawIndexedPrimitive(pt,0,0,5,0,3);

	if(GetAsyncKeyState(VK_SPACE))
	{
		while(GetAsyncKeyState(VK_SPACE));
		pt = (D3DPRIMITIVETYPE)((DWORD)pt+1);
		if(pt==7)
			pt = (D3DPRIMITIVETYPE)4;
		FillBuffers();
	}
	TCHAR buffer[100];
	sprintf(buffer,"Cambiare tipo di primitiva con lo SPAZIO\n"
					"Sto usando %s",GetPT(pt));
	g_font->DrawText(buffer,-1,&g_font_rect,DT_BOTTOM,0xff000000);	
	//Fine della scena
	g_device->EndScene();
	//Mettiamo il tutto a schermo
	g_device->Present(0,0,0,0);
}