// -------------------------------------------------------------
// Vertex Blending
// scritto da Stefano Cristiano (stefano.cristiano@email.it)
//
// Questo esempio si propone di spiegare il funzionamento 
// del vertex blending con la fixed function pipeline
// -------------------------------------------------------------
#include "stdafx.h"
#include <d3dx9.h>
#include <atlbase.h>
#include <vector>
#include <string>


//Questo  un metodo per linkare le librerie
//di DX9
#pragma comment(lib,"d3d9.lib")
#pragma comment(lib,"d3dx9.lib")

//Typedefiamo questi "smart refcounted pointers"
typedef CComPtr<IDirect3D9>			Direct3D9_Ptr;
typedef CComPtr<IDirect3DDevice9>	Direct3DDevice9_Ptr;
typedef CComPtr<ID3DXFont>			D3DXFont_Ptr;
typedef CComPtr<ID3DXMesh>			D3DXMesh_Ptr;
typedef CComPtr<IDirect3DTexture9>	Direct3DTexture9_Ptr;
typedef CComPtr<IDirect3DTexture9>	Direct3DTexture9_Ptr;
typedef CComPtr<ID3DXBuffer>		D3DXBuffer_Ptr;

HWND				g_hwnd;		//La finestra di rendering		
Direct3D9_Ptr		g_d3d;		//L'oggetto IDirect3D (serve x creare il device)
Direct3DDevice9_Ptr g_device;	//Il Device (il nostro acceleratore)
int					g_width;	//Larghezza corrente della finestra
int					g_height;	//Altezza corrente della finestra
bool				g_terminated;//Il programma e' stato terminato?	
RECT				g_font_rect;
std::vector<Direct3DTexture9_Ptr> g_textures;
std::vector<D3DXMATERIAL>		g_materials;
D3DXMesh_Ptr		g_mesh;		//La mesh
D3DXFont_Ptr		g_font;

bool CreateDeviceObjects();
void Draw();
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);


int APIENTRY WinMain(HINSTANCE instance,HINSTANCE,LPSTR,int)
{
	//Registriamo la WNDCLASS
	const WNDCLASS wndclass = {
			0, WndProc, 0,0, instance,0,
			LoadCursor(NULL, IDC_CROSS),
			0,0,_T("VertexBlendingClass")
	};
	g_width = 640;
	g_height= 480;
	g_terminated = false;
	if(!RegisterClass(&wndclass))
		return -1;

	//Creiamo una normalissima finestra
	g_hwnd = CreateWindow(	_T("VertexBlendingClass"),
							_T("Vertex Blending! Tendendo premuto SPAZIO disegna "
								"i triangoli.Premere ESC per uscire."),
							WS_OVERLAPPEDWINDOW,CW_USEDEFAULT,CW_USEDEFAULT,
							g_width,g_height,GetDesktopWindow(),0,instance,0);
	if(!g_hwnd)
		return -1;

	//Creiamo l'oggetto IDirect3D9 (da  notare
	//che assegnamo direttamente il puntatore ricevuto
	//al membro .p dello smart pointer...il perch
	//e' spiegato nell'articolo che spiega il COM)
	g_d3d.p = Direct3DCreate9(D3D_SDK_VERSION);
	if(!g_d3d)
		return -1;

	//Prendiamo la modalit di display corrente del 
	//desktop
	D3DDISPLAYMODE display;
	g_d3d->GetAdapterDisplayMode(D3DADAPTER_DEFAULT,&display);

	//Impostiamo la struttura per un device in finestra,
	//con un solo backbuffer,senza multisampling (antialiasing),
	//e abilitiamo anche lo zbuffer (D3DFMT_D16)
	D3DPRESENT_PARAMETERS params;
	ZeroMemory(&params,sizeof(params));
	params.Windowed					= TRUE;
	params.hDeviceWindow			= g_hwnd;
	params.BackBufferCount			= 1;
	params.BackBufferFormat			= display.Format;
	params.BackBufferWidth			= 640;	
	params.BackBufferHeight			= 480;	
	params.SwapEffect				= D3DSWAPEFFECT_DISCARD;
	params.MultiSampleType			= D3DMULTISAMPLE_NONE;
	params.EnableAutoDepthStencil	= TRUE;
	params.AutoDepthStencilFormat	= D3DFMT_D16;


	//Creiamo il device con accelereazione (D3DDEVTYPE_HAL)
	//usando il software vertex processing (disabilitiamo il T&L
	//della scheda praticamente)
	HRESULT hr;
	hr = g_d3d->CreateDevice(	D3DADAPTER_DEFAULT,D3DDEVTYPE_HAL,
								g_hwnd,D3DCREATE_SOFTWARE_VERTEXPROCESSING,
								&params,&g_device.p);
	if(FAILED(hr))
		return -1;

	//Mostriamo la finestra
	ShowWindow(g_hwnd,SW_SHOW);

	//Aggiorniamo la finestra
	UpdateWindow(g_hwnd);

	//Ne prendiamo le dimensioni
	GetClientRect(g_hwnd,&g_font_rect);

	if(!CreateDeviceObjects())
	{
		//Fine del programma
		MessageBox(g_hwnd,"Errore","Errore di inizializzazione",MB_OK);
		DestroyWindow(g_hwnd);
		return -1;
	}
	//Pompa dei messaggi
	while(!g_terminated)
	{	
		Draw();
		//Smistiamo i messaggi in coda...
		MSG msg;
		while(PeekMessage(&msg,0,0,0,PM_REMOVE))
		{	
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	
	//Fine del programma
	DestroyWindow(g_hwnd);
	return 0;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{

	LRESULT res = 0;
	if(message==WM_CLOSE||(message==WM_KEYDOWN&&wParam==VK_ESCAPE))
	{
		g_terminated = true;
		PostQuitMessage(0);
		return 0;
	}
	res = DefWindowProc(hWnd,message,wParam,lParam);
	return res;
}

struct CUSTOMVERTEX
{
    D3DXVECTOR3 posizione;	//Posizione
    float       blend;		//Peso di blending
    D3DXVECTOR3 normale;	//Normale
    D3DXVECTOR2 texture;	//Coordinate Texture
};

const DWORD CUSTOMVERTEX_FVF =
    D3DFVF_XYZB1 | D3DFVF_NORMAL | D3DFVF_TEX1;

void AssignBlendWeights()
{
	HRESULT hr;
	CUSTOMVERTEX* v=0;
	hr = g_mesh->LockVertexBuffer(D3DLOCK_NOSYSLOCK,(void**)&v);
	if(FAILED(hr))
		return;
	D3DXVECTOR3 bmin,bmax;
	D3DXComputeBoundingBox((D3DXVECTOR3*)v,g_mesh->GetNumVertices(),
							sizeof(CUSTOMVERTEX),&bmin,&bmax);
	float xsize = bmax.x - bmin.x;
	for(int i =0; i < g_mesh->GetNumVertices(); i++)
		v[i].blend = sinf(D3DX_PI*(v[i].posizione.x-bmin.x)/xsize);

	g_mesh->UnlockVertexBuffer();
}

bool LoadMesh(const std::string filename, D3DXMesh_Ptr& mesh, 
				std::vector<D3DXMATERIAL>& materials,
				std::vector<Direct3DTexture9_Ptr>& textures)
{
	DWORD dwNumMaterials;
	HRESULT hr;
	D3DXBuffer_Ptr pD3DXMtrlBuffer;//Per ottenere i materiali
	hr = D3DXLoadMeshFromX( 
			filename.c_str(),		//Nome del file
			D3DXMESH_MANAGED,		//Si gestisce da solo in caso di "Lost Device"
			g_device, NULL,			//Il device e l'adiacenza che non ci interessa
			&pD3DXMtrlBuffer,NULL,	//Ritorna array di materiali (D3DXMATERIAL)
			&dwNumMaterials,		//Ritorna il numero dei material presenti
			&mesh); 				//Ritorna la mesh

	if(FAILED(hr))
		return false;

	//Estraiamo il path dove si trova il file .x
	std::string path = filename.substr(0,filename.find_last_of(_T("\\"))+1);	
	D3DXMATERIAL* d3dxMaterials = (D3DXMATERIAL*)pD3DXMtrlBuffer->GetBufferPointer();		
	for(DWORD i=0; i<dwNumMaterials; i++ )
	{	
		materials.push_back(d3dxMaterials[i]);
		if(!d3dxMaterials[i].pTextureFilename)
		{	
			textures.push_back(0);
			continue;
		}
		std::string texName =d3dxMaterials[i].pTextureFilename;
		Direct3DTexture9_Ptr tex;
		//Proviamo il nome assoluto del file
		hr = D3DXCreateTextureFromFile(g_device,texName.c_str(),&tex);
		if(FAILED(hr))
		{
			//cerchiamo nella dir corrente
			int start = texName.find_last_of('\\');
			texName = texName.substr(start+1,texName.size()-start-1);
			hr = D3DXCreateTextureFromFile(g_device,texName.c_str(),&tex);
			if(FAILED(hr))
			{
				//Cerchiamo nel path del file .x
				texName = path + texName;
				D3DXCreateTextureFromFile(g_device,texName.c_str(),&tex);
			}
		}
		textures.push_back(tex);
	}
	return true;
}

D3DXMATRIX matWorld1,matWorld2;
bool CreateDeviceObjects()
{
	HRESULT hr;
	//Creiamo il font
	D3DXCreateFont(g_device,(HFONT)GetStockObject(SYSTEM_FONT),&g_font.p);
	//creiamo la texture della mesh
	D3DXMesh_Ptr tmpMesh;
	if(!LoadMesh(_T("ASD.X"),tmpMesh,
		g_materials,g_textures))
		return false;
	hr = tmpMesh->CloneMeshFVF(	D3DXMESH_MANAGED,
								CUSTOMVERTEX_FVF,
								g_device,&g_mesh);
	if(FAILED(hr))
		return false;
	AssignBlendWeights();
	//Abilitiamo il dithering, lo zbuffer ed il calcolo
	//dell'illuminazione
	g_device->SetRenderState(D3DRS_DITHERENABLE, TRUE );
	g_device->SetRenderState(D3DRS_ZENABLE, TRUE );
	g_device->SetRenderState(D3DRS_LIGHTING,TRUE);

	//D3D deve disegnare i poligoni la cui normale 
	//e' rivolta verso la camera, usando il verso antiorario
	//dei vertici x determinare dove punta
	g_device->SetRenderState(D3DRS_CULLMODE,D3DCULL_CCW);

	//Moduliam la texture con il colore dell'illuminazione
	g_device->SetTextureStageState(0,D3DTSS_COLOROP,D3DTOP_MODULATE);
	g_device->SetTextureStageState(0,D3DTSS_COLORARG1,D3DTA_TEXTURE);
	g_device->SetTextureStageState(0,D3DTSS_COLORARG2,D3DTA_CURRENT);

	//Abilitiamo il filtering
	g_device->SetSamplerState(0,D3DSAMP_MINFILTER,D3DTEXF_LINEAR);
	g_device->SetSamplerState(0,D3DSAMP_MAGFILTER,D3DTEXF_LINEAR);
	g_device->SetSamplerState(0,D3DSAMP_MIPFILTER,D3DTEXF_LINEAR);

	//Disabilitiamo gli stages successivi
	g_device->SetTextureStageState(1,D3DTSS_COLOROP,D3DTOP_DISABLE);
	g_device->SetTextureStageState(1,D3DTSS_ALPHAOP,D3DTOP_DISABLE);

	//Impostiamo la luce
	D3DLIGHT9 theLight;
	ZeroMemory(&theLight,sizeof(theLight));
	theLight.Type = D3DLIGHT_DIRECTIONAL;
	theLight.Direction = D3DXVECTOR3(1, 1, 1);
	theLight.Ambient  = D3DXCOLOR(0.1f,0.1f,0.1f,1.0f);
	theLight.Diffuse  = D3DXCOLOR(1,1,1,1.0f);
	g_device->SetLight(0,&theLight);
	g_device->LightEnable(0,TRUE);

	//Impostiamo il materiale
	D3DMATERIAL9 m;
	ZeroMemory(&m,sizeof(m));
	m.Ambient = D3DXCOLOR(0.1f,0.1f,0.1f,1.0f);
	m.Diffuse = D3DXCOLOR(1,1,1,1);
	g_device->SetMaterial(&m);	
	g_device->SetRenderState(D3DRS_SPECULARENABLE, FALSE);

	D3DXMatrixIdentity(&matWorld1);
	D3DXMatrixIdentity(&matWorld2);
	return true;
}

void Draw()
{
	//Il nostro parametro temporale
	float t = (float)GetTickCount()/1000.0f;
	//Inizio del frame corrente
	g_device->BeginScene();
	//Puliamo lo schermo con 1 bel nero
	g_device->Clear(0,0,D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,0,1,0);
	if(GetAsyncKeyState(VK_SPACE))
		g_device->SetRenderState(D3DRS_FILLMODE,D3DFILL_WIREFRAME);
	else
		g_device->SetRenderState(D3DRS_FILLMODE,D3DFILL_SOLID);
		
	//Questo sarebbe il rapporto tra le dimensioni dello schermo
	float aspect = (float)g_width/(float)g_height;
	D3DXMATRIX matProj,matView;
	//La matrice di proiezione
	D3DXMatrixPerspectiveFovLH(&matProj,D3DX_PI/3.0f,aspect,0.1f,50.0f);
	//La matrice camera (Fissa)
	D3DXMatrixLookAtLH(&matView,
						&D3DXVECTOR3(3,5,-10),
						&D3DXVECTOR3(0,0,0),
						&D3DXVECTOR3(0,1,0));

	g_device->SetTransform(D3DTS_VIEW,&matView);
	g_device->SetTransform(D3DTS_PROJECTION,&matProj);

	D3DXVECTOR3 axis(
					2.0f+sinf(3.3f*t),
					2.0f+sinf(1.3f*t),
					sinf(2.4f*t));
	static int choice;
	for(int k =0; k < 4; k++)
		if(GetAsyncKeyState('1'+k))
			choice = k;
		
	//Diamo qualche trasformazione a caso alle matrici
	if(choice==0)
	    D3DXMatrixRotationAxis(&matWorld1,&axis, sinf(3*t));
	if(choice==1)
		D3DXMatrixTranslation(&matWorld1,0,0,sinf(2*t)*4);
	if(choice==2)
		D3DXMatrixTranslation(&matWorld2,0,0,sinf(2*t)*4);
	if(choice==3)
	{
		D3DXMatrixTranslation(&matWorld2,0,0,-sinf(2*t)*4);
		D3DXMatrixTranslation(&matWorld1,0,0,sinf(2*t)*4);
	}
	
	//Impostiamo le 2 bones
	g_device->SetTransform(D3DTS_WORLD,&matWorld1);
	g_device->SetTransform(D3DTS_WORLD1,&matWorld2);
	//Attiviamo il vertex blending a 2 bones
	g_device->SetRenderState(D3DRS_VERTEXBLEND,D3DVBF_1WEIGHTS);		

	for(UINT i = 0; i < g_materials.size(); i++)
	{
		//Impostiamo il materiale del subset
		g_device->SetMaterial(&g_materials[i].MatD3D);
		//Impostiamo la texture del subset
		g_device->SetTexture(0,g_textures[i]);
		//Disegnamo la mesh
		g_mesh->DrawSubset(i);
	}
	
	//Disegnamo del testo bianco (0xffffffff)
	g_font->DrawText(	_T(	"Vertex Blending!!\n"
							"Cambiare le varie animazioni delle bones con "
							"i tasti \'1\',\'2\',\'3\',\'4\'\n"
							"Premere SPAZIO per visualizzare in Wireframe"),-1,
							&g_font_rect,0,0xffffffff);
	//Fine della scena
	g_device->EndScene();
	//Mettiamo il tutto a schermo
	g_device->Present(0,0,0,0);
}