// File: MySnakeMIDlet.java

// Pacchetti di J2ME necessari alla classe.
import javax.microedition.lcdui.*;
import javax.microedition.midlet.*;

// MySnakeMIDlet estende MIDlet e implementa CommandListener.
public class MySnakeMIDlet extends MIDlet implements CommandListener {
	
	// Queste costanti identificano i vari dispositivi.
	public static final int COLORS_DEVICE     =  0; // Schermo a colori.
	public static final int GRAYTONES_DEVICE  =  1; // Schermo a toni di grigio.
	public static final int BW_DEVICE         =  2; // Schermo in bianco e nero.
	
	// Qui sar memorizzato il tipo di dispositivo in uso, in corrispondenza
	// delle tre differenti costanti sopra dichiarate.
	private int deviceType;
	
	// La canvas che conterr e mostrer la schermata di avvio del gioco.
	private MySnakeTitleCanvas titleCanvas = null;
	
	// Questa canvas conterr il gioco vero e proprio.
	private MySnakeGameCanvas gameCanvas = null;
	
	// I comandi ammessi dalla MIDlet per l'interazione con l'utente.
	private Command cmdExit  = new Command("Esci", Command.EXIT, 1);
	private Command cmdStart = new Command("Gioca", Command.SCREEN, 2);
	private Command cmdBack  = new Command("Indietro", Command.BACK, 2);
	
	// Costruttore. Prepara gli elementi di base per l'avvio del gioco.
	public MySnakeMIDlet() {
		// Acquisisce informazioni sul dispositivo in uso.
		Display display = Display.getDisplay(this);
		if (display.isColor()) {
			deviceType = COLORS_DEVICE;
		} else if (display.numColors() >= 256) {
			deviceType = GRAYTONES_DEVICE;
		} else {
			deviceType = BW_DEVICE;
		}
		// Crea la canvas da visualizzare all'avvio.
		titleCanvas = new MySnakeTitleCanvas(deviceType);
		// Aggiunge i comandi "Esci" e "Gioca".
		titleCanvas.addCommand(cmdExit);
		titleCanvas.addCommand(cmdStart);
		// Imposta il gestore di eventi per i comandi.
		// Vedere il metodo commandAction di questa stessa classe.
		titleCanvas.setCommandListener(this);
	}
	
	// Questo metodo gestisce i comandi azionati dall'utente.
	public void commandAction(Command c, Displayable d) {
		if (c == cmdExit) {
			// Chiusura della MIDlet.
			shutdown();
		} else if (c == cmdStart) {
			// Avvio di una nuova partita.
			if (gameCanvas == null) {
				gameCanvas = new MySnakeGameCanvas(deviceType);
				gameCanvas.addCommand(cmdExit);
				gameCanvas.addCommand(cmdBack);
				gameCanvas.setCommandListener(this);
			}
			Display display = Display.getDisplay(this);
			display.setCurrent(gameCanvas);
			gameCanvas.startGame();
		} else if (c == cmdBack) {
			// Torna alla schermata del titolo, se possibile.
			if (gameCanvas.stopGame()) {
				Display display = Display.getDisplay(this);
				display.setCurrent(titleCanvas);
			}
		}
	}
	
	// Questo metodo viene eseguito non appena la MIDlet  avviata.
	public void startApp() {
		Display display = Display.getDisplay(this);
		display.setCurrent(titleCanvas);
	}
	
	// Metodi la cui implementazione  richiesta dalla classe MIDlet.
	// In questo caso sono vuoti perch non si intende usufruire delle loro
	// funzionalit.
	public void pauseApp() {}
	public void destroyApp(boolean b) {}
	
	// Termina l'esecuzione della MIDlet.
	public void shutdown() {
		destroyApp(false);
		notifyDestroyed();
	}
	
}